;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "python-mode" "20260410.1849"
  "Python major mode."
  ()
  :url "https://gitlab.com/groups/python-mode-devs"
  :commit "c6f7bcd79d43b2a4915b75c481355601f95968a0"
  :revdesc "c6f7bcd79d43"
  :keywords '("python" "languages" "oop")
  :maintainers '((nil . "python-mode@python.org")))
